#include "config.h"
#include <dgesvd.h>
#include <dgesvd.h>
#include <dgesdd.h>
#include <dgelss.h>
#include <dgglse.h>
#include <dgeqr2.h>
//#include <lmdif.h> //from minpack
//#include <lmder.h> //from minpack
#include <minpack.h>
#include <cminpack.h>
#include <dsyev.h>

#include "matrix.h"
#include "vec3.h"

NACB_BEGIN

int Matrix::gcount = 0;
int Matrix::acount = 0;


Matrix Matrix::matrix2rod(){
  if(m!=3||n!=3){
    printf("error in matrix2rod\n");
    return Matrix(0,0);
  }
  Matrix axis(3,1);
  double angle;
  double ct;
  double st;
  double t1;
  double t2;
  double t3;
  ct=(get(0,0)+get(1,1)+get(2,2)-1)/2;
  t1=get(2,1)-get(1,2);
  t2=get(0,2)-get(2,0);
  t3=get(1,0)-get(0,1);
  st=0.5*::sqrt(t1*t1+t2*t2+t3*t3);

  angle=atan2(st,ct);
  double temp=1.0/(2*st);
  axis[0]=temp*t1;
  axis[1]=temp*t2;
  axis[2]=temp*t3;
    
  axis*=angle;
    
#ifdef PD
  axis.printMatlab("axis");
  printMatlab("matrix");
#endif
  return axis;
}

Matrix Matrix::rod2matrix(){
#ifdef MATRIX_CHECK
  if(m!=3||n!=1){
    printf("error \n");
  }
#endif
  Vec3d axis(get(0,0),get(1,0),get(2,0));
  double angle = axis.normalize();
  Matrix mat = Matrix::eye(3);
  Matrix S(3,3);
  S(0,0)=0;
  S(0,1)=-axis.z;
  S(0,2)= axis.y;
  S(1,0)= axis.z;
  S(1,1)=0;
  S(1,2)=-axis.x;
  S(2,0)=-axis.y;
  S(2,1)= axis.x;
  S(2,2)=0;
  Matrix S2=S*S;
  S*=(sin(angle));
  S2*=(1-cos(angle));
  
  mat+=S;
  mat+=S2;
  return mat;
}

//Implmentation from the matlab rot2yxz
Matrix Matrix::rot2zxy(){
  double ax,ay,az;
  double sy,cy;
  double eps=10e-16;
  if(fabs((*this)(2,0)<eps) && fabs((*this)(2,2))<eps){
    ay = 0;
    az = atan2((*this)(1,0),(*this)(0,0));
    ax = atan2((*this)(2,1),(*this)(2,2));
  }
  else{
    ay = atan2(-(*this)(2,0),(*this)(2,2));
    sy = sin(ay);
    cy = cos(ay);
    ax = atan2((*this)(2,1),-sy*(*this)(2,0)+cy*(*this)(2,2));
    az = atan2(cy*(*this)(1,0)+sy*(*this)(1,2),cy*(*this)(0,0)+sy*(*this)(0,2));
  }
  Matrix ret(3,1);
  ret[0]=az;ret[1]=ax;ret[2]=ay;
  return ret;
}

Matrix Matrix::rot2xyz(){
  if(m!=3 || n!=3){
    return Matrix(0,0);
  }
  Matrix xyz(3,1);
  double thetaX,thetaZ,thetaY = asin(get(0,2));
  if(thetaY < M_PI/2){
    if(thetaY > -M_PI/2){
      thetaX=atan2(-get(1,2),get(2,2));
      thetaZ=atan2(-get(0,1),get(0,0));
    }
    else{
      thetaX=-atan2(get(1,0),get(1,1));
      thetaZ=0;
    }
  }
  else{
    thetaX=atan2(get(1,0),get(1,1));
    thetaZ=0;
  }
  xyz[0]=thetaX;
  xyz[1]=thetaY;
  xyz[2]=thetaZ;
  return xyz;
}

void Matrix::eigSym(Matrix & eigVectors,Matrix & eigenValues,int toget){
  if(m!=n)throw size_mismatch();
  Matrix D=transpose();  
  Matrix work(1,1);
  int lwork=-1;
  int info;

  eigenValues=Matrix(m,1);
  dsyev_('V','U', m, D.data, m, eigenValues.data, work.data, lwork, info);
  
  lwork=int(work[0]);
  work=Matrix(lwork,1);

  dsyev_('V','U', m, D.data, m, eigenValues.data, work.data, lwork, info);
  eigVectors=D.transpose();
}

void Matrix::svd(Matrix & a,Matrix & w,Matrix & v){
  Matrix t=transpose();
  w=Matrix(std::min(m,n),1);
  Matrix U(m,m);
  v=Matrix(n,n);
  int worksize=2*std::max(3*std::min(m,n)+std::max(m,n),5*std::min(m,n));
  Matrix work(worksize,1);
  int info=0;
  dgesvd_('A','A',m,n,t.data,m,w.data,U.data,U.m,v.data,v.m,
	  work.data,worksize,info);
  if(info!=0)fprintf(stderr,"error in Lsvd (returned %d)\n",info);
  a=U.transpose();
}
void Matrix::svdLarge(Matrix & a,Matrix & w,Matrix & v){
  Matrix t=transpose();
  w=Matrix(std::min(m,n),1);
  Matrix U(m,m);
  v=Matrix(n,n);
  int worksize=std::min(m,n)*std::min(m,n) +
    std::max(std::max(m,n),4*std::max(m,n)*std::max(m,n)+4*std::max(m,n));
  int * iwork=new int[8*std::min(m,n)];
  worksize*=2;
  Matrix work(worksize,1);
  int info=0;
  dgesdd_('A',m,n,t.data,m,w.data,U.data,U.m,v.data,v.m,work.data,worksize,iwork,info);
  if(info!=0)fprintf(stderr,"error in LsvdLarge (returned %d)\n",info);
  a=U.transpose();
  delete [] iwork;
}

//Lapack qr decomposition
//returns 1 on success
int Matrix::qr(Matrix & Q,Matrix & R){
  Matrix T=this->transpose();//Make column major for lapack
  int info=0;
  Matrix tau(std::min(m,n),1);
  Matrix work(n,1);
  dgeqr2_(m,n,T.data,m,tau.data,work.data,info);
  if(info<0){
    printf("argument %d to dgeqr2 was bad\n",-info);
    return 0;
  }
  T=T.transpose();//Undo the column major for me
  Q=Matrix::eye(m,m);
  R=Matrix(m,n);
  R.setAll(0);
  for(int i=0;i<std::min(m,n);i++)
    for(int j=i;j<n;j++)
      R(i,j)=T(i,j);
  //Unpack Q, its a product of reflectors (according to lapack)
  Matrix v(Q.m,1);
  for(int i=0;i<std::min(m,n);i++){
    for(int j=0;j<i;j++)
      v[j]=0;
    v[i]=1;
    for(int j=i+1;j<Q.m;j++)
      v[j]=T(j,i);
    Matrix H=Matrix::eye(Q.m,Q.m);
    H-=(v*v.transpose()*tau[i]);
    Q=Q*H;
  }
  return info==0;
}

double Matrix::det() const{
  if(m!=n){
    printf("can you get det of non-square matrix\n");
    return 0;
  }
  if(m==1||n==1)return data[0];
  //expand about top row
  double ret=0;
  int sign=1;
  Matrix subMatrix(m-1,n-1);
    
  for(int j=0;j<n;j++){
    for(int i=1;i<m;i++){
      int ki=0;
      for(int k=0;k<n;k++){
	if(k==j)
	  continue;
	subMatrix(i-1,ki)=get(i,k);
	ki++;
      }
    }
    ret+=get(0,j)*sign*subMatrix.det();
    sign*=-1;
  }
  return ret;
}

Matrix Matrix::inverse() const {
  if(m!=n){
    printf("trying to invert non-square matrix\n");
    return Matrix();
  }
  int i, j, k;
    
  Matrix cop = copy();
  Matrix inv = Matrix::eye(m);
  for(j=0;j<m;j++){
    double largest=0;
    int largei=0;
    for(i=j;i<m;i++){
      if(fabs(cop(i,j))>largest){
	//printf("setting %f %f\n",largest,cop->get(i,j));
	largei=i;
	largest=fabs(cop(i,j));
      }
    }
    // printf("largest: %f\n",largest);
    //pivot
    for(k=0;k<m;k++){
      double temp=cop(j,k);
      cop(j,k)=cop(largei,k);
      cop(largei,k)=temp;
	
      temp=inv(j,k);
      inv(j,k)=inv(largei,k);
      inv(largei,k)=temp;
    }
      
    double ival=1.0/cop(j,j);
    for(k=0;k<m;k++){
      cop(j,k)=cop(j,k)*ival;
      inv(j,k)=inv(j,k)*ival;
      //data[j+4*k]*=ival;
      //inv[j+4*k]*=ival;
    }
    for(i=0;i<m;i++){
      if(i!=j){
	double scale=cop(i,j);
	for(k=0;k<m;k++){
	  cop(i,k)=cop(i,k)-scale*cop(j,k);
	  inv(i,k)=inv(i,k)-scale*inv(j,k);
	}
      }
    }
  }
  return inv;
}


void Matrix::print(){
  for(int i=0;i<m;i++){
    for(int j=0;j<n;j++){
      printf("%f ",(*this)(i,j));
    }
    printf("\n");
  }
}

int Matrix::lsqnonlin(int (* func)(void * ptr,
				   int  m, 
				   int  n,
				   const double * x,
				   double * fvec,
				   double * fjac,
				   int ldfjac,
				   int iflag),
		      void * ptr,
                      Matrix & x,Matrix & fvec,
                      double ftol,double xtol,double gtol,
                      int maxfev,double * diag,
                      double factor,int nprint)
{
  int mode=1;
  int n=x.m*x.n;      //Number of variables                                                                                                                 
  int m=fvec.m*fvec.n;//Number of functions                                                                                                                 
  int info=0;
  int nfev=0;
  int njev=0;
  Matrix fjac(m,n);
  int ldfjac=m;
  int * ipvt=new int[m*n];
  Matrix qtf(n,1);//output containing q'*fvec                                                                                                               
  Matrix wa1(n,1);//work arrays                                                                                                                             
  Matrix wa2(n,1);
  Matrix wa3(n,1);
  Matrix wa4(m,1);//work array of length m                                                                                                                  

  // Allocate the diagonal scale factors if the user                                                                                                        
  // did not provide, and set the mode accordingly                                                                                                          
  if(diag)mode=2;
  else diag=new double[n];
  
  lmder((int (*)(void *,int,int, const double*, double*, double*, int, int))func,
	ptr,
	m,n,x.data,fvec.data,fjac.data,ldfjac,ftol,xtol,
	gtol,maxfev,diag,mode,
	factor,nprint,&nfev,&njev,
	ipvt,qtf.data,wa1.data,wa2.data,wa3.data,wa4.data);
  
  if(mode==1)delete [] diag;
  delete [] ipvt;
  return info;
}



int Matrix::lsqnonlin(void (* func)(const int & m,
				    const int & n,
				    double * x,
				    double * fvec,
				    double * fjac,
				    const int & ldfjac,
				    int & iflag),
		      Matrix & x,Matrix & fvec,
		      double ftol,double xtol,double gtol,
		      int maxfev,double * diag,
		      double factor,int nprint)
{
  int mode=1;
  int n=x.m*x.n;      //Number of variables
  int m=fvec.m*fvec.n;//Number of functions
  int info=0;
  int nfev=0;
  int njev=0;
  Matrix fjac(m,n);
  int ldfjac=m;
  int * ipvt=new int[m*n];
  Matrix qtf(n,1);//output containing q'*fvec
  Matrix wa1(n,1);//work arrays
  Matrix wa2(n,1);
  Matrix wa3(n,1);
  Matrix wa4(m,1);//work array of length m
      
  // Allocate the diagonal scale factors if the user 
  // did not provide, and set the mode accordingly
  if(diag)mode=2;
  else diag=new double[n];
      
  lmder_((void (*)(const int *m, const int *n, const double *x, double *fvec,
		   double *fjec, const int *ldfjac, int *iflag ))func,
	 &m,&n,x.data,fvec.data,fjac.data,&ldfjac,&ftol,&xtol,
	 &gtol,&maxfev,diag,&mode,
	 &factor,&nprint,&info,&nfev,&njev,
	 ipvt,qtf.data,wa1.data,wa2.data,wa3.data,wa4.data);
      
  if(mode==1)delete [] diag;
  delete [] ipvt;
  return info;
}   

int  Matrix::lsqnonlin(int (* func)(void * ptr,
				    int  m,
				    int  n,
				    const double * x,
				    double * fvec,
				    int iflag),
		       void * ptr,
                       Matrix & x,Matrix & fvec,
                       double ftol,double xtol,double gtol,
                       int maxfev,double epsfcn,double * diag,
                       double factor,int nprint)
{
  int mode=1;
  int n=x.m*x.n;      //Number of variables                                                                                                                 
  int m=fvec.m*fvec.n;//Number of functions                                                                                                                 
  int info=0;
  int nfev=0;
  Matrix fjac(m,n);
  int ldfjac=m;
  int * ipvt=new int[m*n];
  Matrix qtf(n,1);//output containing q'*fvec                                                                                                               
  Matrix wa1(n,1);//work arrays                                                                                                                             
  Matrix wa2(n,1);
  Matrix wa3(n,1);
  Matrix wa4(m,1);//work array of length m                                                                                                                  

  // Allocate the diagonal scale factors if the user                                                                                                        
  // did not provide, and set the mode accordingly                                                                                                          
  if(diag)mode=2;
  else diag=new double[n];

  lmdif(func,ptr,m,n,x.data,fvec.data,ftol,xtol,
	gtol,maxfev,epsfcn,diag,mode,
	factor,nprint,&nfev,fjac.data,ldfjac,
	ipvt,qtf.data,wa1.data,wa2.data,wa3.data,wa4.data);

  if(mode==1)delete [] diag;
  delete [] ipvt;
  return info;
}

int  Matrix::lsqnonlin(void (* func)(const int & m,
				     const int & n,
				     double * x,
				     double * fvec,
				     int & iflag),
		       Matrix & x,Matrix & fvec,
		       double ftol,double xtol,double gtol,
		       int maxfev,double epsfcn,double * diag,
		       double factor,int nprint)
{
  int mode=1;
  int n=x.m*x.n;      //Number of variables
  int m=fvec.m*fvec.n;//Number of functions
  int info=0;
  int nfev=0;
  Matrix fjac(m,n);
  int ldfjac=m;
  int * ipvt=new int[m*n];
  Matrix qtf(n,1);//output containing q'*fvec
  Matrix wa1(n,1);//work arrays
  Matrix wa2(n,1);
  Matrix wa3(n,1);
  Matrix wa4(m,1);//work array of length m
  
  // Allocate the diagonal scale factors if the user 
  // did not provide, and set the mode accordingly
  if(diag)mode=2;
  else diag=new double[n];
  
  lmdif_((void(*)(const int *,const int *,const double *,double *,int *))func,
	  &m,&n,x.data,fvec.data,&ftol,&xtol,
	  &gtol,&maxfev,&epsfcn,diag,&mode,
	  &factor,&nprint,&info,&nfev,fjac.data,&ldfjac,
	  ipvt,qtf.data,wa1.data,wa2.data,wa3.data,wa4.data);
  
  if(mode==1)delete [] diag;
  delete [] ipvt;
  return info;
}

Matrix Matrix::linLeastSq(Matrix &A,Matrix &b){
  Matrix s(std::min(A.m,A.n),1);
  int rank;
  int info;
  int lwork=2*(3*std::min(A.m,A.n)+std::max(2*std::min(A.m,A.n),std::max(std::max(A.m,A.n),b.n )));
  Matrix work(lwork,1);
  Matrix At=A.transpose();
  Matrix bt=b.transpose();
  dgelss_(A.m,A.n,b.n,At.data,A.m,bt.data,b.m,s.data,-1,rank,work.data,lwork,info);
  //dgelsx_(A.m,A.n,b.
  Matrix x(A.n,1);
  for(int i=0;i<x.m;i++){
    x[i]=bt[i];
  }
  return x;
}


Matrix Matrix::linLeastSq(Matrix &A,Matrix &b,Matrix & B,Matrix & d)
{
  Matrix At=A.transpose();
  Matrix Bt=B.transpose();
  Matrix bt=b.transpose();//Just do so data isn't clobbered
  Matrix dt=d.transpose();//dido
  Matrix X(A.n,1);
  int lwork=-1;
  Matrix work(1,1);
  int info=0;
  dgglse_(A.m,A.n,B.m,At.data,A.m,Bt.data,
	  B.m,bt.data,dt.data,X.data,work.data,lwork,info);
  lwork=(int)(work[0]);
  work=Matrix(lwork,1);
  dgglse_(A.m,A.n,B.m,At.data,A.m,Bt.data,
	  B.m,bt.data,dt.data,X.data,work.data,lwork,info);
  return X;
}

NACB_END


#ifdef TEST_MATRIX
int main()
{
  Matrix A(2,2);
  A(0,0)=2;
  A(1,1)=1;
  A.print();
  
  Matrix B=A,C;
  B=A;
  B=A;
  printf("before\n");
  A.printMatlab("A");
  B.printMatlab("B");
  B=B+A;
  A.printMatlab("A");
  B.printMatlab("B");
  B=A*B+A*B+B.transpose()*A.transpose();
  B.printMatlab("B");

  Matrix D(2,2);
  D(0,0)=2;
  D(1,0)=1;
  D(0,1)=1;
  D(1,1)=1;
  printf("About to do svd\n");
  D.svd(A,B,C);
  A.print();
  B.print();
  C.print();
  D.det();
  printf("done svd\n");

  Matrix R=Matrix::random(3,3);
  
  R.printMatlab("R");
  printf("Determinant %f\n",R.det());

}

#endif
