#include "quaternion.h"

NACB_BEGIN

std::ostream & operator<<(std::ostream & os,const Quaternion & q){
  q.print(os);
  return os;
}

void Quaternion::getMatrix(Mat3x3 & mat)const{
  mat(0,0)=(1.0-2.0*(v.y*v.y+v.z*v.z));
  mat(0,1)=2.0*(v.x*v.y-a*v.z);
  mat(0,2)=2.0*(v.x*v.z+a*v.y);
  mat(1,0)=2.0*(v.x*v.y+a*v.z);
  mat(1,1)=1.0-2.0*(v.x*v.x+v.z*v.z);
  mat(1,2)=2.0*(v.y*v.z-a*v.x);
  mat(2,0)=2.0*(v.x*v.z-a*v.y);
  mat(2,1)=2.0*(v.y*v.z+a*v.x);
  mat(2,2)=1.0-2.0*(v.x*v.x+v.y*v.y);
}

Quaternion Quaternion::fromMatrixRows(const Vec3d & r1,
				      const Vec3d & r2,
				      const Vec3d & r3){
  double trace=(r1.x+r2.y+r3.z+1);
  /*double qx2=fabs( r1.x-r2.y-r3.z+1)/4;
    double qy2=fabs(-r1.x+r2.y-r3.z+1)/4;
    double qz2=fabs(-r1.x-r2.y+r3.z+1)/4;
    double qw2=fabs(trace/4);

    double m=std::max(std::max(qx2,qy2),std::max(qz2,qw2));
  */
    
  Quaternion q;
  double qw,qx,qy,qz,s;
  double sx=(1.0+r1.x-r2.y-r3.z)*4;
  double sy=(1.0-r1.x+r2.y-r3.z)*4;
  double sz=(1.0-r1.x-r2.y+r3.z)*4;
  //printf("trace:%f  %f,%f,%f\n",trace,sx,sy,sz);
    
  double m=std::max(std::max(sx,sy),std::max(trace/4,sz));
    
  //was >0.1 and it was fucked then...spent almost an entire day tracking that fucker down
  //this code should be improved
  if(trace>1e-6 ||(trace/4>=m)){
    qw=sqrt(trace)/2;
    qx=(r3.y-r2.z)/(4*qw);
    qy=(r1.z-r3.x)/(4*qw);
    qz=(r2.x-r1.y)/(4*qw);
  }
  //from www.euclideanspace.com/maths/geometry
  else if(r1.x>r2.y && r1.x>r3.z){
    s=sqrt(sx);
    qx=0.25*s;
    qy=(r1.y+r2.x)/s;
    qz=(r1.z+r3.x)/s;
    qw=(r2.z-r3.y)/s;
    //printf("using x\n");
  }
  else if(r2.y>r3.z){
    s=sqrt(sy);
    qx=(r1.y+r2.x)/s;
    qy=0.25*s;
    qz=(r2.z+r3.y)/s;
    qw=(r1.z-r3.x)/s;
    //printf("using y\n");
  }
  else{
    s=sqrt(sz);
    qx=(r1.z+r3.x)/s;
    qy=(r2.z+r3.y)/s;
    qz=0.25*s;
    qw=(r1.y-r2.x)/s;
    //printf("using z\n");
  }
  q.a=qw;
  q.v=Vec3d(qx,qy,qz);
  q.normalize();
  //printf("length in track %f\n",l);
  return q;
}

//Some derivatives
Quaternion Quaternion::rod_dx(const Vec3d & v){
  double len=sqrt(v.x*v.x+v.y*v.y+v.z*v.z);
  double dlen_dx=v.x/len;
  
  double a;
  Vec3d vo;
  a=-sin(len/2.0)*dlen_dx/2.0;
  
  if(len>1e-10){
    double ca=cos(len/2.0);
    double sa=sin(len/2.0);
    
    vo.x=sa/len+v.x*((ca/2)*dlen_dx*len-sa*dlen_dx)/(len*len);
    vo.y=v.y*((ca/2)*dlen_dx*len-sa*dlen_dx)/(len*len);
    vo.z=v.z*((ca/2)*dlen_dx*len-sa*dlen_dx)/(len*len);
  }
  else{
    //near x==0, len ~= x
    //and sin(len/2) ~= x/2 ==> sin(len/2)/len ~=0.5
    a=0;
    vo.x=vo.y=vo.z=0;
    vo.x=0.5;
  }
  return Quaternion(a,vo);
};

//Some derivatives
Quaternion Quaternion::rod_dy(const Vec3d & v){
  double len=sqrt(v.x*v.x+v.y*v.y+v.z*v.z);
  double dlen_dy=v.y/len;
  
  double a;
  Vec3d vo;
  a=-sin(len/2.0)*dlen_dy/2.0;
  
  if(len>1e-10){
    double ca=cos(len/2.0);
    double sa=sin(len/2.0);
    
    vo.x=v.x*((ca/2)*dlen_dy*len-sa*dlen_dy)/(len*len);
    vo.y=sa/len+v.y*((ca/2)*dlen_dy*len-sa*dlen_dy)/(len*len);
    vo.z=v.z*((ca/2)*dlen_dy*len-sa*dlen_dy)/(len*len);
  }
  else{
    //near x==0, len ~= x
    //and sin(len/2) ~= x/2 ==> sin(len/2)/len ~=0.5
    a=0;
    vo.x=vo.y=vo.z=0;
    vo.y=0.5;
  }
  return Quaternion(a,vo);
};

Quaternion Quaternion::rod_dz(const Vec3d & v){
  double len=sqrt(v.x*v.x+v.y*v.y+v.z*v.z);
  double dlen_dz=v.z/len;
  
  double a;
  Vec3d vo;
  a=-sin(len/2.0)*dlen_dz/2.0;
  
  if(len>1e-10){
    double ca=cos(len/2.0);
    double sa=sin(len/2.0);
    
    vo.x=v.x*((ca/2)*dlen_dz*len-sa*dlen_dz)/(len*len);
    vo.y=v.y*((ca/2)*dlen_dz*len-sa*dlen_dz)/(len*len);
    vo.z=sa/len+v.z*((ca/2)*dlen_dz*len-sa*dlen_dz)/(len*len);
  }
  else{
    //near x==0, len ~= x
    //and sin(len/2) ~= x/2 ==> sin(len/2)/len ~=0.5
    a=0;
    vo.x=vo.y=vo.z=0;
    vo.z=0.5;
  }
  return Quaternion(a,vo);
};

void Quaternion::toEuler(const char * order,double angles[3]) const{
  double e=((order[2]+1)%3==order[1])?1:-1;
  angles[0]=atan2(2.0*(a*v.data[(int)order[0]]-e*v.data[(int)order[1]]*v.data[(int)order[2]]),
		  1.0-2.0*(v.data[(int)order[0]]*v.data[(int)order[0]]+
			   v.data[(int)order[1]]*v.data[(int)order[1]]));
  angles[1]=asin(2.0*(a*v.data[(int)order[1]]+e*v.data[(int)order[0]]*v.data[(int)order[2]]));
    
  if(fabs(angles[1]-M_PI/2.0)<3e-8 ||
     fabs(angles[1]+M_PI/2.0)<3e-8){
    angles[2]=0;
    angles[0]=atan2(v.data[(int)order[0]],a);
  }
  else{
    angles[2]=atan2(2*(a*v.data[(int)order[2]]-e*v.data[(int)order[0]]*v.data[(int)order[1]]),
		    1.0-2.0*(v.data[(int)order[1]]*v.data[(int)order[1]]+
			     v.data[(int)order[2]]*v.data[(int)order[2]]));
  }
}

NACB_END
