#ifndef SPARSE_MATRIX_H
#define SPARSE_MATRIX_H

#include "nspace.h"
#include "matrix.h"
#include <vector>
#include <algorithm>

NACB_BEGIN

class SparseMatrix{
 public:
  int m;
  int n;
  int nzmax;
  int nnz;
  double * data;
 protected:
  int * row;
  int * col;

 public:
  SparseMatrix(int m,int n,int nzmax);
  SparseMatrix(const Matrix & mat);
  SparseMatrix(const SparseMatrix & sm);
  SparseMatrix(int m,std::vector<std::vector<std::pair<int,double> > > & cols);

  void ensure(int nzmax);
  double & insert_no_check(int i,int j,const double & d=0);
  double & get(int i,int j);
  int have(int i,int j) const{
    int end=col[j+1];
    for(int ind=col[j];ind<end;ind++){
      if(row[ind]==i)return 1;
    }
    return 0;
  }
  void operator=(const SparseMatrix & sm);
  void printMatlab(char * str);
  
  double & operator()(int i,int j){
    return get(i,j);
  }
  void release(){
    delete [] row;
    delete [] data;
    delete [] col;
    nnz=nzmax=n=m=0;
    row = 0;
    data = 0;
    col = 0;
  }
  void lsqr_vec_prod_add(long trans,double * X,double * Y);
  void matprod(int trans,double * X,double * Y);
  Matrix svd(double & sing_val,int smallest);
  Matrix svd_last(double & sing_val){
    return svd(sing_val,1);
  }
  Matrix lsqr(Matrix & b,double xtol=1e-6,int maxits=200);
  Matrix linLeastSq(Matrix & b,double xtol=1e-6,int maxits=200);
  Matrix gaussSeidel(Matrix & b,double xtol=1e-6,int maxits=200);

  static Matrix umfsolve(SparseMatrix & A,Matrix & b);
  ~SparseMatrix(){
    release();
  }
};

NACB_END

#endif
