#include "sparsematrix.h"
#include "matrix.h"
#include <sys/time.h>

using namespace nacb;

void check(Matrix & A,Matrix & b,
	   Matrix & x1,Matrix & x2,const char * nm){

  Matrix diff = x1-x2;
  Matrix tmp = A*x1-b;
  double r1 = tmp.dot(tmp);
  tmp = A*x2-b;
  double r2 = tmp.dot(tmp);

  printf("%s diff: %e (%e,%e) residuals %e,%e\n",nm,sqrt(diff.dot(diff)),
	 sqrt(x1.dot(x1)),sqrt(x2.dot(x2)),r1,r2);

}

double elapsed(){
  static struct timeval st,et,dt;
  gettimeofday(&et,0);

  timersub(&et,&st,&dt);
  st=et;

  return (double(dt.tv_sec)+double(dt.tv_usec)/1e6);
}

int main(int ac,char * av[]){
  int m = ac>1?atoi(av[1]):1000;
  Matrix A=Matrix::random(m,m);
  Matrix b=Matrix::random(m,1);
  
  for(int i=0; i<A.m; i++){
    for(int j=i+1;j<A.n; j++){
      if(rand()%100<95){
	A(i,j)=A(j,i)=0;
      }
    }
  }
  A=A+Matrix::eye(A.m,A.n)*10;

  printf("solving for x1...\n");
  elapsed();
  Matrix x1=Matrix::LlinLeastSq(A,b);
  printf("...got x1 took %f\n",elapsed());
  
  Matrix diff;
  SparseMatrix sA(A);
#ifdef USE_UMFPACK
  printf("solving for x2...\n");
  Matrix x2=SparseMatrix::umfsolve(sA,b);
  printf("...got x2 took %f\n",elapsed());
  check(A,b,x1,x2,"umfpack");
#endif

  elapsed();
  Matrix x3=sA.linLeastSq(b,1e-8,1000);
  printf("...got x3 took %f\n",elapsed());
    
  check(A,b,x1,x3,"linLeastSq");
  
  printf("..doing gaussSeidel\n");elapsed();
  Matrix x4=sA.gaussSeidel(b,1e-6,500);
  printf("took %f\n",elapsed());
  check(A,b,x1,x4,"gaussSeidel");

  printf("..doing lsqr\n");elapsed();
  Matrix x5=sA.lsqr(b,1e-6,500);
  printf("took %f\n",elapsed());
  check(A,b,x1,x5,"lsqr");
  return 0;
}
