#include "vec4.h"

#include <iostream>

//class Vec4<double>;
//class Vec4<float>;

NACB_BEGIN

//namespace nacb{

template <class T>
void Vec4<T>::rotx(double t){
  double c=cos(t),s=sin(t);
  T yn=c*y-s*z;
  T zn=s*y+c*z;
  y=yn;
  z=zn;
}
template <class T>
void Vec4<T>::roty(double t){
  double c=cos(t),s=sin(t);
  T xn= c*x+s*z;
  T zn=-s*x+c*z;
  x=xn;
  z=zn;
}
template <class T>
void Vec4<T>::rotz(double t){
  double c=cos(t),s=sin(t);
  T xn=c*x-s*y;
  T yn=s*x+c*y;
  x=xn;
  y=yn;
}
template <class T>
void Vec4<T>::dx(double t){
  double c=-sin(t),s=cos(t);
  T yn=c*y-s*z;
  T zn=s*y+c*z;
  y=yn;
  z=zn;
  w=x=0;
}
template <class T>
void Vec4<T>::dy(double t){
  double c=-sin(t),s=cos(t);
  T xn= c*x+s*z;
  T zn=-s*x+c*z;
  x=xn;
  z=zn;
  w=y=0;
}
template <class T>
void Vec4<T>::dz(double t){
  double c=-sin(t),s=cos(t);
  T xn=c*x-s*y;
  T yn=s*x+c*y;
  x=xn;
  y=yn;
  w=z=0;
}

VEC4_INSTANTIATIONS

NACB_END


#ifdef TEST_VEC4
using namespace std;
using namespace nacb;

int test_assert(bool val,const char * message){
  if(!val){
    std::cout << "failed: " << message << "\n";
    return 0;
  }
  return 1;
}
int main(int ac,char * av[]){
  Vec4d d(0,1,0);
  Vec4d dorig=d;
  std::cout << "d:"<< d <<"\n";

  double eps=1e-10;
  int succ=0;
  succ+=test_assert((d-dorig).len()<eps,"copy constructor failed");
  succ+=test_assert((d-Vec4d(d.x,d.y,d.z,d.w)).len()<eps,"constructor failed");

  d.rotatex( M_PI/2);
  d.rotatex(-M_PI/2);

  succ+=test_assert((d-dorig).len()<eps,"rotation x failed");

  d=dorig;
  d.rotatey( M_PI/2);
  d.rotatey(-M_PI/2);

  succ+=test_assert((d-dorig).len()<eps,"rotation y failed");

  d=dorig;
  d.rotatez( M_PI/2);
  d.rotatez(-M_PI/2);

  succ+=test_assert((d-dorig).len()<eps,"rotation y failed");

  d=dorig;
  d.rotatez( M_PI/2);
  d.rotatex( M_PI/4);
  d.rotatey( M_PI/8);
  d.rotatey(-M_PI/8);
  d.rotatex(-M_PI/4);
  d.rotatez(-M_PI/2);

  succ+=test_assert((d-dorig).len()<eps,"rotation z,x,y failed");
  
  Vec4d x(1,0,0,0),y(0,1,0,0),z(0,0,1,0);
  succ+=test_assert(((x^y)-z).len()<eps,"x^y!=z");
  succ+=test_assert(((y^z)-x).len()<eps,"y^z!=x");

  x=Vec4d(0.4,0.2,0.11,0.9);
  succ+=test_assert(((x*1.1)/1.1-x).len()<eps,"* / ");

  double a=1.5;
  Vec4d res=(z*a*1.2-a*x/a+y/a+a*x/a-(1.0/a)*y-a*1.2*z);
  succ+=test_assert(res.len()<eps,"add and subtract");

  res=z;
  res*=a*1.2;
  res-=(a*x/a);
  res+=(y/a);
  res-=(1.0/a)*y;
  res+=a*x/a;
  res-=a*1.2*z;
  succ+=test_assert(res.len()<eps,"add and subtract");

  Vec4f rf(0.3,0.4,0.5,1);
  Vec4d rd=(Vec4d)rf;
  (Vec4d)rf;
  succ+=test_assert(((Vec4f)rd-((Vec4d)rf)).len()<eps,"vec4 cast");

  std::cout << "Successful tests: "<<succ<< "\n";

  return 0;
}
#endif
